#include <libxml/globals.h>
#include <libxml/xinclude.h>
#include <libxml/HTMLparser.h>
#include <libxml/HTMLtree.h>
#include <libxml/xpathInternals.h>
#include <libxml/uri.h>
#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/extensions.h>
#include <libxslt/xsltutils.h>
#include <libexslt/exslt.h>
#include <sysexits.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>
#include <unistd.h>
#include <time.h>
#include "config.h"

#define WK_XSLT_EXT_NAMESPACE ((const xmlChar *) "http://xml.utmc.or.jp/wk")

void rewrite_links_in_tree(xmlNodePtr root, xmlChar* base,
			   int (*compare)(const char* a, const char *b));
xmlChar* convert_absuri_to_reluri(xmlChar *absuri, xmlChar* baseuri);
void traverse_tree(xmlNodePtr children, int indent, xmlChar *base,
		   int (*compare)(const char* a, const char *b));
void set_parameter(char *key, size_t keylen, char *val,
	      const char ***params, int *param_idx, int *params_size);
static void
wk_xslt_ext_file_mtime_function(xmlXPathParserContextPtr ctxt, int nargs);

typedef enum {
	IFT_XML,
	IFT_HTML
} InputFileType;

const char* directory_index[] = {
	"index.html",
	"index.htm",
	NULL
};

void
usage(const char* argv0)
{
    const char *p = strrchr(argv0, '/');
    p = p ? p + 1 : argv0;
    fprintf(stderr,
"%s version " VERSION "\n"
"usage: %s [-p param=val ...] [-XU] STYLE XML URI\n"
"       %s [-p param=val ...] [-XU] HTML URI\n"
"\n"
"       -p : pass parameter to XSLT.\n"
"       -X : Disable exslt.\n"
"       -U : Disable URL rewriting..\n"
, p, p, p);
}

#define PARAMS_SIZE_STEP     10

int
main(int argc, char* argv[])
{
    char *path_param;
    xsltStylesheetPtr style;
    xmlDocPtr doc, res;
    xmlNodePtr root;
    char *stylefile, *inputfile, *baseuri;
    InputFileType filetype = IFT_XML;
    int (*compare)(const char* a, const char* b);
    const char **params;
    int c, param_idx, params_size;
    extern char *optarg;
    extern int optind, opterr, optopt;
    char *p;
    int use_exslt = 1;
    int rewrite_url = 1;

    param_idx = 0;
    params = NULL;
    params_size = 0;
    while ((c = getopt(argc, argv, "p:XU")) != -1) {
	switch (c) {
	case 'p':
	    if (p = strchr(optarg, '=')) {
		set_parameter(optarg, p - optarg, p + 1,
			      &params, &param_idx, &params_size);
	    }
	    break;
	case 'X':
	    use_exslt = 0;
	    break;
	case 'U':
	    rewrite_url = 0;
	    break;
	default:
	    usage(argv[0]);
	    exit(EX_USAGE);
	}
    }

    if (argc - optind < 3 || argc - optind > 4) {
	usage(argv[0]);
	exit(1);
    }

    if (argc - optind == 3) {
	stylefile = argv[optind];
	inputfile = argv[optind + 1];
	baseuri = argv[optind + 2];
	filetype = IFT_XML;
	compare = strcmp;
    } else {
	inputfile = argv[optind];
	baseuri = argv[optind];
	filetype = IFT_HTML;
	compare = strcasecmp;
    }

    /* Check URI parameter and initialize parameter for XSLT */
    if (*baseuri != '/') {
	fprintf(stderr, "Error: URI must be started with '/'.\n");
	exit(1);
    }
    path_param = xmlMalloc(strlen(baseuri) + 3);
    if (!path_param) {
	fprintf(stderr, "Error: Failed to allocate memory for parameter\n");
	exit(1);
    }
    sprintf(path_param, "'%s'", baseuri);
    set_parameter("path", 4, path_param, &params, &param_idx, &params_size);

    params[param_idx] = NULL;

    /* Initialize XML parser */
    LIBXML_TEST_VERSION;
    xmlInitMemory();
    xmlSubstituteEntitiesDefault(1);
    if (getenv("http_proxy")) {
	xmlNanoHTTPScanProxy(getenv("http_proxy"));
    }
    if (use_exslt) {
	exsltRegisterAll();
    }
    xsltRegisterExtModuleFunction((const xmlChar *) "file-mtime",
				  WK_XSLT_EXT_NAMESPACE,
				  wk_xslt_ext_file_mtime_function);
    xsltRegisterTestModule();
    xmlLoadExtDtdDefaultValue = XML_DETECT_IDS | XML_COMPLETE_ATTRS;

    if (filetype == IFT_XML) {
	/* Parse XML file */
	doc = xmlParseFile(inputfile);
	xmlXIncludeProcess(doc);

	/* Parse style sheet */
	style = xsltParseStylesheetFile((const xmlChar *)stylefile);
	res = xsltApplyStylesheet(style, doc, params);
    } else {
	/* Parse HTML file */
	res = htmlParseFile(inputfile, NULL);
    }

    /* Rewrite links */
    root = xmlDocGetRootElement(res);
    if (rewrite_url && root && root->children) {
	rewrite_links_in_tree(root, baseuri, compare);
    }

    if (filetype == IFT_XML) {
	/* Output results (XML) */
	xsltSaveResultToFile(stdout, res, style);

	/* Finalize XSLT/XML */
	xsltFreeStylesheet(style);
	xmlFreeDoc(doc);
    } else {
	/* Output results (HTML) */
	htmlDocDump(stdout, res);
    }
    xmlFreeDoc(res);
    xsltCleanupGlobals();
    xmlCleanupParser();

    exit(0);
}

void
rewrite_links_in_tree(xmlNodePtr root, xmlChar* base,
		      int (*compare)(const char *a, const char *b))
{
    traverse_tree(root->children, 0, base, compare);
}

xmlChar*
convert_absuri_to_reluri(xmlChar *absuri, xmlChar* baseuri)
{
    unsigned char *a, *b, *last_slash, *tmp;
    xmlChar* ret;
    size_t retsize;
    int depth, i;

    a = (unsigned char *)absuri;
    b = (unsigned char *)baseuri;
    if (*absuri != '/' || *baseuri != '/') return NULL;
#if 0
    /* remove /../ and /./ */
    resolve_uri(absuri);
    resolve_uri(baseuri);
#endif
    last_slash = b;
    while (*a == *b && *a && *b) {
	if (*b == '/') last_slash = b;
	a++;
	b++;
    }
    depth = 0;
    tmp = last_slash + 1;
    while (*tmp) {
	if (*tmp == '/') depth++;
	tmp++;
    }
    tmp = absuri + (last_slash - baseuri) + 1;
    retsize =  strlen(tmp) + 3 * depth + 1;
    ret = xmlMalloc(retsize);
    if (ret == NULL) return NULL;
    *ret = '\0';
    for (i = 0; i < depth; i++) {
	strcat(ret, "../");
    }
    strcat(ret, tmp);
    return ret;
}

typedef struct tagLinkElement {
    const char* tag_name;
    const char* attr_name;
} LinkElement;

LinkElement links[] = {
    { "a", "href" },
    { "link", "href" },
    { "img", "src" },
    { "form", "action" },
    { NULL, NULL }
};

void
traverse_tree(xmlNodePtr children, int indent, xmlChar *base,
	      int (*compare)(const char *a, const char *b))
{
    xmlNodePtr node;
    for (node = children; node; node = node->next) {
	if (node->type == XML_ELEMENT_NODE) {
	    xmlChar *href;
	    LinkElement *l = links;
	    while (l->tag_name) {
		if (compare((char *)node->name, l->tag_name) != 0) {
		    l++;
		    continue;
		}
		href = xmlGetProp(node, l->attr_name);
		if (href && href[0] == '/' && href[1] != '/') {
		    xmlChar* newuri;
		    if ((newuri = convert_absuri_to_reluri(href, base)) != NULL) {
			xmlSetProp(node, l->attr_name, newuri);
			xmlFree(newuri);
		    }
		}
		xmlFree(href);
		break;
	    }
	    if (node->children)
		traverse_tree(node->children, indent + 1, base, compare);
	}
    }
}

void*
alloc_memory(void* ptr, size_t size)
{
    void* ret;
    ret = xmlRealloc(ptr, size);
    if (ret == NULL) {
	if (errno == ENOMEM) {
	    fprintf(stderr, "Error: Can not allocate memory.\n");
	} else {
	    fprintf(stderr, "Error.\n");
	}
	exit(1);
    }
    return ret;
}

void
set_parameter(char *key, size_t keylen, char *val,
	      const char ***params, int *param_idx, int *params_size)
{
    char *s;

    if (*param_idx >= *params_size) {
	    char **new_params;
	    int new_params_size = *params_size + PARAMS_SIZE_STEP;
	    *params = alloc_memory(*params, new_params_size * sizeof(char *));
    }

    s = alloc_memory(NULL, keylen + 1);
    memcpy(s, key, keylen);
    s[keylen] = '\0';
    (*params)[*param_idx++] = s;

    s = alloc_memory(NULL, strlen(val) + 3);
    sprintf(s, "'%s'", val);
    (*params)[*param_idx++] = s;
}

static void
wk_xslt_ext_file_mtime_function(xmlXPathParserContextPtr ctxt, int nargs)
{
    xmlChar *href = NULL;
    xmlChar *ret = NULL;
    xmlChar *base = NULL;
    xmlChar *target = NULL;
    struct stat st;
    time_t t;
    struct tm *mtime;
    char retstr[32];
    int f, h, m;
    /* extern long timezone; (defined in time.h) */

    if (nargs  != 1) {
	xmlXPathSetArityError(ctxt);
	return;
    }

    href = xmlXPathPopString(ctxt);
    base = xmlNodeGetBase(ctxt->context->doc, ctxt->context->node);
    target = xmlBuildURI(href, base);

    if (stat(target, &st) < 0) {
	/* no such file or directory */
	xmlXPathReturnEmptyString(ctxt);
	goto error;
    } else if (S_ISREG(st.st_mode)) {
	/* regular file */
	/* do nothing */
	;
    } else if (S_ISDIR(st.st_mode)) {
	/* directory */
	char cwd[MAXPATHLEN];
	int found = 0;

	if (getcwd(cwd, sizeof(cwd))) {
	    const char** dp = NULL;
	    DIR *dh;
	    struct dirent *ent;

	    chdir(target);
	    for (dp = directory_index; *dp != NULL; dp++) {
		if (stat(*dp, &st) == 0) {
		    found = 1;
		    break;
		}
	    }
	    if (!found && (dh = opendir("."))) {
		while (!found && (ent = readdir(dh))) {
		    if (stat(ent->d_name, &st) < 0 || !S_ISREG(st.st_mode))
			continue;
		    for (dp = directory_index; *dp != NULL; dp++) {
			size_t dplen = strlen(*dp);
			if (strncmp(ent->d_name, *dp, dplen) != 0)
			    continue;
			if (ent->d_name[dplen] == '.') {
			    found = 1;
			    break;
			}
		    }
		}
		closedir(dh);
	    }
	    chdir(cwd);
	}
	if (!found || !S_ISREG(st.st_mode)) {
	    xmlXPathReturnEmptyString(ctxt);
	    goto error;
	}
    } else {
	xmlXPathReturnEmptyString(ctxt);
	goto error;
    }

    t = st.st_mtime;
    mtime = localtime(&t);

    f = (timezone < 0) ? -1 : 1;
    if (timezone < 0)
	timezone = -timezone;
    h = (int)(timezone / 3600);
    m = (int)((timezone - (h * 3600)) / 60);
    sprintf(retstr, "%04d-%02d-%02dT%02d:%02d:%02d%c%02d:%02d",
	    mtime->tm_year + 1900, mtime->tm_mon + 1, mtime->tm_mday,
	    mtime->tm_hour, mtime->tm_min, mtime->tm_sec,
	    (f < 0 ? '+' : '-'), h, m);
    
    xmlXPathReturnString(ctxt, xmlStrdup(retstr));

error:
    if (href != NULL)
	xmlFree(href);
    if (base != NULL)
	xmlFree(base);
    if (target != NULL)
	xmlFree(target);
}

/* end of wk */
